from __future__ import annotations
from royalnet.typing import *
import threading
import redis


__all__ = (
    "Baron",
    "BaronError",
    "BaronAlreadyStartedError",
    "BaronListenerThread",
)


class Baron:
    """The Baron module connects to a Redis database to send and receive messages."""

    def __init__(self,
                 redis_args: Mapping[str, Any],):
        self.publisher: redis.Redis = redis.Redis(**redis_args)
        self.listen_thread: BaronListenerThread = BaronListenerThread(publisher=self.publisher)
        self.is_started = False

    def listener(self) -> redis.client.PubSub:
        """Get the listener of the Baron module."""
        return self.listen_thread.listener

    def start(self) -> Baron:
        """Start the listen thread of the Baron module."""
        if self.is_started:
            raise BaronAlreadyStartedError("This Baron module was already started somewhere else.")
        self.listen_thread.start()
        self.is_started = True
        return self


class BaronError(Exception):
    """An error of the Baron module."""


class BaronAlreadyStartedError(Exception):
    """This Baron module was already started somewhere else."""


class BaronListenerThread(threading.Thread):
    """A Thread that creates a PubSub from a Redis instance and constantly listens to it.

    It ignores all messages that do not have an associated callback."""
    def __init__(self, publisher: redis.Redis, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.listener: redis.client.PubSub = publisher.pubsub()

    def run(self) -> None:
        while True:
            self.listener.listen()



from royalnet.typing import *
import abc

__all__ = (
    "AsyncChallenge",
    "TrueAsyncChallenge",
)


class AsyncChallenge(metaclass=abc.ABCMeta):
    """A filter for inputs passed to an AsyncCampaign."""

    @abc.abstractmethod
    async def filter(self, data: Any) -> bool:
        """Decide if the data should be skipped or not."""
        raise NotImplementedError()


class TrueAsyncChallenge(AsyncChallenge):
    """An AsyncChallenge which always returns True."""

    async def filter(self, data: Any) -> bool:
        return True



from ..errors import RoyalnetException


class ScrollException(RoyalnetException):
    """An exception raised by the Scroll module."""


class NotFoundError(ScrollException):
    """The requested config key was not found."""


class InvalidFormatError(ScrollException):
    """The requested config key is not valid."""


class ParseError(ScrollException):
    """The config value could not be parsed correctly."""


class InvalidFileType(ParseError):
    """The type of the specified config file is not currently supported."""


__all__ = (
    "ScrollException",
    "NotFoundError",
    "InvalidFormatError",
    "ParseError",
    "InvalidFileType",
)



from royalnet.typing import *
import abc

__all__ = (
    "Challenge",
    "TrueChallenge",
)


class Challenge(metaclass=abc.ABCMeta):
    """A filter for inputs passed to a Campaign."""

    @abc.abstractmethod
    def filter(self, data: Any) -> bool:
        """Decide if the data should be skipped or not."""
        raise NotImplementedError()


class TrueChallenge(Challenge):
    """A Challenge which always returns True."""

    async def filter(self, data: Any) -> bool:
        return True



from royalnet.typing import *
import sqlalchemy as sa
import sqlalchemy.orm as saorm


__all__ = (
    "Alchemist",
)


class Alchemist:
    """The Alchemist module connects to a relational database with SQLAlchemy."""

    def __init__(self,
                 engine_args: Iterable[Any],
                 engine_kwargs: Mapping[str, Any]):
        self.engine: sa.engine.Engine = sa.create_engine(*engine_args, **engine_kwargs)
        self.Session: Union[saorm.sessionmaker, Type[saorm.Session]] = saorm.sessionmaker(bind=self.engine)

    def add_metadata(self, metadata: sa.MetaData):
        """Bind a MetaData object to the engine, and create all tables linked with it."""
        metadata.bind = self.engine
        metadata.create_all()



"""This package adds some new common types to the default typing package.

It should be imported with: ::

    from royalnet.typing import *

"""

from typing import *

JSONScalar = Union[
    None,
    float,
    int,
    str,
]

JSON = Union[
    JSONScalar,
    List["JSON"],
    Dict[str, "JSON"],
]

Adventure = Generator[
    Union[
        Tuple[
            Optional["Challenge"],
            ...
        ],
        "Adventure",
        "Challenge",
        None,
    ],
    Any,
    Any,
]

AsyncAdventure = AsyncGenerator[
    Union[
        Tuple[
            Optional["AsyncChallenge"],
            ...
        ],
        "AsyncAdventure",
        "AsyncChallenge",
        None,
    ],
    Any,
]



from royalnet.typing import *
import os
import json
import re
import toml
import json

from .errors import *


class Scroll:
    """A configuration handler that allows getting values from both the environment variables and a config file."""

    key_validator = re.compile(r"^[A-Z.]+$")

    loaders = {
        ".json": json.load,
        ".toml": toml.load
    }

    def __init__(self, namespace: str, config: Optional[Dict[str, JSON]] = None):
        self.namespace: str = namespace
        self.config: Optional[Dict[str, JSON]] = config

    @classmethod
    def from_toml(cls, namespace: str, file_path: os.PathLike):
        with open(file_path) as file:
            config = toml.load(file)
        return cls(namespace, config)

    @classmethod
    def from_json(cls, namespace: str, file_path: os.PathLike):
        with open(file_path) as file:
            config = json.load(file)
        return cls(namespace, config)

    @classmethod
    def from_file(cls, namespace: str, file_path: os.PathLike):
        file, ext = os.path.splitext(file_path)
        lext = ext.lower()

        with open(file_path) as file:
            try:
                config = cls.loaders[lext](file)
            except KeyError:
                raise InvalidFileType(f"Invalid extension: {lext}")

        return cls(namespace, config)

    @classmethod
    def _validate_key(cls, item: str):
        check = cls.key_validator.match(item)
        if not check:
            raise InvalidFormatError()

    def _get_from_environ(self, item: str) -> JSONScalar:
        """Get a configuration value from the environment variables."""
        key = f"{self.namespace}_{item.replace('.', '_')}"

        try:
            j = os.environ[key]
        except KeyError:
            raise NotFoundError(f"'{key}' was not found in the environment variables.")

        try:
            value = json.loads(j)
        except json.JSONDecodeError:
            raise ParseError(f"'{key}' contains invalid JSON: {j}")

        return value

    def _get_from_config(self, item: str) -> JSONScalar:
        """Get a configuration value from the configuration file."""
        if self.config is None:
            raise NotFoundError("No config file has been loaded.")

        chain = item.split(".")

        current = self.config

        for element in chain:
            try:
                current = current[element]
            except KeyError:
                raise NotFoundError(f"'{item}' was unreachable: could not find '{element}'")

        return current

    def __getattribute__(self, item: str):
        self._validate_key(item)
        try:
            return self._get_from_environ(item)
        except NotFoundError:
            return self._get_from_config(item)


__all__ = ("Scroll",)



import pytest
from ..campaign import Campaign
from ..challenge import Challenge, TrueChallenge
from ..exc import *


def test_creation():
    def gen():
        yield

    Campaign.create(start=gen())


def test_send_receive():
    def gen():
        ping = yield
        assert ping == "Ping!"
        yield None, "Pong!"

    campaign = Campaign.create(start=gen())
    pong, = campaign.next("Ping!")
    assert pong == "Pong!"


class FalseChallenge(Challenge):
    def filter(self, data) -> bool:
        return False


def test_failing_check():
    def gen():
        yield

    campaign = Campaign.create(start=gen(), challenge=FalseChallenge())
    with pytest.raises(ChallengeFailedError):
        campaign.next()


def test_switching():
    def gen_1():
        yield
        yield gen_2()

    def gen_2():
        yield
        yield None, "Second message!"
        yield None

    campaign = Campaign.create(start=gen_1())
    data, = campaign.next()
    assert data == "Second message!"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['royalnet',
 'royalnet.alchemist',
 'royalnet.baron',
 'royalnet.campaigns',
 'royalnet.campaigns.tests',
 'royalnet.scroll',
 'royalnet.typing']

package_data = \
{'': ['*']}

install_requires = \
['hiredis>=1.1.0,<2.0.0',
 'redis>=3.5.3,<4.0.0',
 'sqlalchemy>=1.3.19,<2.0.0',
 'toml>=0.10.1,<0.11.0']

setup_kwargs = {
    'name': 'royalnet',
    'version': '6.0.0a2',
    'description': 'A multipurpose bot and web framework',
    'long_description': None,
    'author': 'Stefano Pigozzi',
    'author_email': 'ste.pigozzi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
